/* rainbow.c -- rainbow visualization plugin for xmms 
   Written by Bryan Rittmeyer <bryanr@caltech.edu>
   Copyright (C) 1999 Horizon Technologies
   http://horizon.pair.com/                          */

#include "rainbow.h"

/* external plugin functions */
static void rainbow_init(void);
static void rainbow_cleanup(void);
static void rainbow_playback_start(void);
static void rainbow_playback_stop(void);
static void rainbow_render_pcm(gint16 data[2][512]);
static void rainbow_about(void);

/* Define plugin */
VisPlugin rainbow_vp =
{
  NULL,
  NULL,
  0,
  "Rainbow",
  1,
  0,
  rainbow_init,
  rainbow_cleanup,
  rainbow_about,
  plugin_configure,
  NULL,
  rainbow_playback_start,
  rainbow_playback_stop,
  rainbow_render_pcm,
  NULL
};

/* return visplugin info for xmms */
VisPlugin *get_vplugin_info(void)
{
  return &rainbow_vp;
}

void plugin_read_config(void)
{
	ConfigFile *cfg;	
	gchar *filename;

	plugin_cfg.tdfx_mode = FALSE;
	
	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	cfg = xmms_cfg_open_file(filename);
	
	if (cfg)
	{
		xmms_cfg_read_boolean(cfg, "Rainbow", "tdfx_fullscreen",
				      &plugin_cfg.tdfx_mode);		
		xmms_cfg_free(cfg);
	}
	g_free(filename);
}

/* initialize plugin */
static void rainbow_init(void)
{
  newframe=FALSE;

  plugin_read_config();
  if(!plugin_cfg.tdfx_mode)
    start_display();

/*   seconds=time(NULL); */
/*   srand(seconds); */
}

/* shut down plugin */
static void rainbow_cleanup(void)
{
  stop_display();
}

/* begin playback */
static void rainbow_playback_start(void)
{
/*   create_rainbow(); */
 /*  going=TRUE; */
  if(plugin_cfg.tdfx_mode) {
    if(window)
      stop_display();
    start_display();
  }
}

/* stop playback */
static void rainbow_playback_stop(void)
{
/*   create_rainbow(); */
/*   going=FALSE; */
  newframe=FALSE;
  if(plugin_cfg.tdfx_mode) {
    stop_display();
  }
}

/* render visualization frame */
static void rainbow_render_pcm(gint16 data[2][512])
{
  int total=0;

  /*  for (y = 0; y < 2; y++) { */
  /*    last=(data[0][0]>>8); */

    for (x = 0; x < 512; x++) {
      total += abs(data[0][x]>>8);
      /*      last = data[0][x]>>8; */
    }	

    ftotal = (float)total / 512.0f;
    ftotal /= 256.0f;
  /*  } */

  if(ftotal > Max - ((Max / 100.0f) * 10.0f))
    Max = ftotal;

  if( (ftotal + ((Max /100.0f) * 10.0f)) >= Max) {
      if(direction == 0)
	direction = 1;
      else
	direction = 0;
  }

  if(Rotation > 90.0f) {
    if(direction == 0)
      direction = 1;
    else
      direction = 0;
  }

  if(Rotation < -90.0f) {
    if(direction == 0)
      direction = 1;
    else
      direction = 0;
  }

  if(direction == 0)
    Rotation += (ftotal * 20.0f);
  
  if(direction == 1)
    Rotation -= (ftotal * 20.0f);

  switch(colourSwap)
    {
    case 0:
      colourSwap = 1;
      glClearColor(Max * 3.0f, 0.0f, 0.0f, 0.0f);
      break;
      
    case 1:
      colourSwap = 2;
      glClearColor(0.0f, Max * 3.0f, 0.0f, 0.0f);
      break;

    case 2:
      colourSwap = 0;
      glClearColor(0.0f, 0.0f, Max * 3.0f, 0.0f);
      break;
    };

  newframe=TRUE;
  //  draw_rainbow();

}

/* about box */
static void rainbow_about(void)
{

  dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog), "About Rainbow");
  gtk_container_border_width(GTK_CONTAINER(dialog), 5);
  label = gtk_label_new("Rainbow Visualization Plugin for xmms\n \
Written by Bryan Rittmeyer <bryanr@caltech.edu>\n \
Copyright (C) 1999 Horizon Technologies\n \
http://horizon.pair.com/\n\n \
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program; if not, write to the Free Software\n\
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,\n\
USA.");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, TRUE, TRUE, 0);
  gtk_widget_show(label);
  
  button = gtk_button_new_with_label(" Close ");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(dialog));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  gtk_widget_show(dialog);
  gtk_widget_grab_focus(button);
}

static Window create_window(int width, int height)
{
  int attr_list[] = { GLX_RGBA,
		      GLX_DEPTH_SIZE, 16, 
		      GLX_DOUBLEBUFFER,
		      None };
  int scrnum;
  XSetWindowAttributes attr;
  unsigned long mask;
  Window root, win;
  XVisualInfo *visinfo;
  Atom wm_protocols[1];
  
  scrnum = DefaultScreen(dpy);
  root = RootWindow(dpy, scrnum);

  visinfo = glXChooseVisual(dpy, scrnum, attr_list);
  if(!visinfo)
     return 0;

  attr.background_pixel = 0;
  attr.border_pixel = 0;
  attr.colormap = colormap = XCreateColormap(dpy, root, visinfo->visual, AllocNone);
  attr.event_mask = StructureNotifyMask | KeyPressMask;
  mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;

  win = XCreateWindow(dpy, root, 0, 0, width, height,
                      0, visinfo->depth, InputOutput,
                      visinfo->visual, mask, &attr);
  XmbSetWMProperties(dpy, win, "Rainbow",
                     "Rainbow", NULL, 0, NULL, NULL,
			   NULL);
  wm_delete_window_atom = wm_protocols[0] = XInternAtom(dpy,
							"WM_DELETE_WINDOW",
							 False);
  XSetWMProtocols(dpy, win, wm_protocols, 1);

  glxcontext = glXCreateContext(dpy, visinfo, NULL, True);

  glXMakeCurrent(dpy, win, glxcontext);

  return win;
}


void *draw_thread_func(void *arg)
{
  Bool configured=FALSE;

#ifdef HAVE_SCHED_SETSCHEDULER
  if(xmms_check_realtime_priority()) {
    struct sched_param sparam;
    sparam.sched_priority = sched_get_priority_max(SCHED_OTHER);
    pthread_setschedparam(pthread_self(), SCHED_OTHER, &sparam);
  }
#endif

  while(going)
    {
      while(XPending(dpy))
	{
	  XEvent event;
	  KeySym keysym;
	  char buf[16];

	  XNextEvent(dpy,&event);
	  switch(event.type)
	    {
	    case ConfigureNotify:
	      ClientW=event.xconfigure.width;
	      ClientH=event.xconfigure.height;
	      glViewport(0,0,event.xconfigure.width,event.xconfigure.height);
	      if(plugin_cfg.tdfx_mode && !grabbed_pointer) {
		XGrabPointer(dpy, window,
			     True, ButtonPressMask,
			     GrabModeAsync,
			     GrabModeAsync,
			     window, None, CurrentTime);
		grabbed_pointer = TRUE;
	      }
	      configured=TRUE;
	      break;
	    case KeyPress:
	      XLookupString(&event.xkey,buf,16,&keysym,NULL);
	      switch(keysym)
		{
		case XK_Escape:
		  GDK_THREADS_ENTER();
		  gtk_idle_add(disable_func,NULL);
		  GDK_THREADS_LEAVE();
		  break;
		case XK_z:
		  xmms_remote_playlist_prev(rainbow_vp.xmms_session);
		case XK_x:
		  xmms_remote_play(rainbow_vp.xmms_session);
		  break;
		case XK_c:
		  xmms_remote_pause(rainbow_vp.xmms_session);
		  break;
		case XK_v:
		  xmms_remote_stop(rainbow_vp.xmms_session);
		  break;
		case XK_b:
		  xmms_remote_playlist_next(rainbow_vp.xmms_session);
		  break;
/* 		case XK_Up: */
/* 		  yrot-=0.1; */
/* 		  if(yrot<-3.0) */
/* 		    yrot=-3.0; */
/* 		  break; */
/* 		case XK_Down: */
/* 		  yrot+=0.1; */
/* 		  if(yrot>3.0) */
/* 		    yrot=3.0; */
/* 		  break; */
/* 		case XK_Left: */
/* 		  xrot-=0.1; */
/* 		  if(xrot<-3.0) */
/* 		    xrot=-3.0; */
/* 		  break; */
/* 		case XK_Right: */
/* 		  xrot+=0.1; */
/* 		  if(xrot>3.0) */
/* 		    xrot=3.0; */
/* 		  break; */
		}
	      break;
	    case ClientMessage: /* Not the X! Anything but that! */
	      if ((Atom)event.xclient.data.l[0] == wm_delete_window_atom) {
		  GDK_THREADS_ENTER();
		  gtk_idle_add(disable_func, NULL);
		  GDK_THREADS_LEAVE();
		}
	      break;
	    }
	}
      if((configured)&(newframe))
      	draw_rainbow();
    }
  pthread_exit(NULL);
}

static gint disable_func(gpointer data)
{
  rainbow_vp.disable_plugin(&rainbow_vp);
  return FALSE;
}

static void start_display(void)
{
  if(plugin_cfg.tdfx_mode)
    setenv("MESA_GLX_FX", "fullscreen", TRUE);
  else
    unsetenv("MESA_GLX_FX");

  dpy=XOpenDisplay(NULL);
  window=create_window(640,480);
  XMapWindow(dpy, window);

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();
  gluPerspective(35.0f, (float)ClientW / (float)ClientH, 1.0f, 8.0f);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  gluLookAt(0.0f, 0.0f, 2.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX, CameraTrans);

  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, ObjectTrans);

  glLineWidth(1.01f);

  glEnable(GL_DEPTH_TEST);
/*   glEnable(GL_ALPHA_TEST); */
/*   glEnable(GL_BLEND); */
/*   glBlendFunc(GL_DST_ALPHA,GL_DST_ALPHA); */

  going=TRUE;
  pthread_create(&draw_thread,NULL,draw_thread_func,NULL);
}

static void stop_display(void)
{
#ifdef HAVE_XMESAFINDBUFFER
  XMesaBuffer xmesabuffer;
#endif

  if(going)
    {
      going = FALSE;
      pthread_join(draw_thread, NULL);
    }

  if(glxcontext)
    {
      glXMakeCurrent(dpy, 0, NULL);
      glXDestroyContext(dpy, glxcontext);
      glxcontext = NULL;
#ifdef HAVE_XMESAFINDBUFFER
      xmesabuffer = XMesaFindBuffer(dpy, window);
      if(xmesabuffer)
	XMesaDestroyBuffer(xmesabuffer);
#endif
    }	
  if(window)
    {
      if(grabbed_pointer)
	{
	  XUngrabPointer(dpy, CurrentTime);
	  grabbed_pointer = FALSE;
	}
      
      XDestroyWindow(dpy, window);
      window = 0;
    }
  if(colormap)
    {
      XFreeColormap(dpy, colormap);
      colormap = 0;
    }
  if(dpy)
    {
      XCloseDisplay(dpy);
      dpy = NULL;
    }
}

static void draw_rainbow(void)
{

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();// load the identity matrix
  glMultMatrixf(CameraTrans);// now multiply the camera trans
  glMultMatrixf(ObjectTrans);// now the object trans

  for(i = 0.25; i <= 2.0f; i += 0.25)
    {
      glRotatef(Rotation + (i / 7.0f), 0.0f, 0.0f, 1.0f);
      
      glBegin(GL_QUADS);
 /*      glColor4f(1.0f, 1.0f, 1.0f, TRANSPARENCY); */
      glColor3f(1.0f, 1.0f, 1.0f);
      glVertex3f(bLEFT, bTOP, ZFACTOR);
/*       glColor4f(0.0f, 1.0f, 0.5f, TRANSPARENCY); */
      glColor3f(0.0f, 1.0f, 0.5f);
      glVertex3f(bLEFT, bTOP + bHEIGHT, ZFACTOR);
      
/*       glColor4f(1.0f, 0.0f, 0.0f, TRANSPARENCY); */
      glColor3f(1.0f, 0.0f, 0.0f);
      glVertex3f(bLEFT + bWIDTH, bTOP + bHEIGHT, ZFACTOR);
/*       glColor4f(0.0f, 0.0f, 1.0f, TRANSPARENCY); */
      glColor3f(0.0f, 0.0f, 1.0f);
      glVertex3f(bLEFT + bWIDTH, bTOP, ZFACTOR);
      glEnd();
    }

  glXSwapBuffers(dpy,window); /* Swap those buffers */
  newframe=FALSE; /* we no longer have a new frame */
}
