/*
 * Fixed-size stack allocation. Read am_alloc.h for details.
 *
 * Michael Pearson <alcaron@ozemail.com.au>
 *
 */

#include<stdio.h>
#include<stdlib.h>
#include"am_alloc.h"
#include"utility.h"

amObject * amNewObject(size_t size, int blocksize)
{
	amObject * obj;

	/* Trying to create an object with a size smaller than amChunk is
       a big no-no! */
	if(size < sizeof(amChunk)) {
		warning("am_alloc: Program attempted to allocate a chunk of"
		        "lower than %d bytes, defaulting to %d.",
		        sizeof(amChunk), sizeof(amChunk));
		size = sizeof(amChunk); 
	}

	/* Allocate the object and set some standard vars */
	obj = (amObject *) safeMalloc(sizeof(amObject));
	obj->size = size;
	obj->blocksize = blocksize;
	obj->freestack = NULL;
	obj->blockstack = NULL;

	/* Create a simple block and set the freestack up */
	amNewBlock(obj);

	/* Return address to new object */
	return obj;
}       

void amNewBlock(amObject *obj)
{
	amChunk * block;
	amBlock * blockdef;
	amChunk * mover;
	amChunk * previous;
	size_t blocks;

	/* Allocate space for the block */
	blocks = obj->size * obj->blocksize;
	block = safeMalloc(blocks);

	/* Set up the stacks in the block, and add to the obj's freestack */
	mover = block;
	while ((size_t) mover < (size_t) block + blocks - obj->size) {
		previous = mover;
		(size_t) mover += obj->size;
		previous->next = mover;         
	}
	mover->next = obj->freestack;
	obj->freestack = block;

	/* Set up the block definition, and add to obk's blockstack */
	blockdef = safeMalloc(sizeof(amBlock));
	blockdef->begin = block;
	blockdef->next = obj->blockstack;
	obj->blockstack = blockdef;
}

void amKillOb(amObject * obj)
{
	amBlock * mover = obj->blockstack;
	amBlock * previous;

	/* Firstly, kill all the blocks */
	while(mover) {
		previous = mover;
		mover = mover->next;
		free(previous->begin);
		free(previous);
	}

	/* Then kill the obdef */
	free(obj);
}

void * amAlloc(amObject *obj)
{
	void * ptr;

	if(!obj->freestack) {
		amNewBlock(obj);
	}

	/* Simple enough. Detach the chunk from the free stack */
	ptr = (void *) obj->freestack;  
	obj->freestack = obj->freestack->next;
	return ptr;
}

void amFree(amObject *obj, void * chunk)
{
	/* Again, simple enough. Attach it to the free stack */
	((amChunk *)chunk)->next = obj->freestack;
	obj->freestack = chunk;
}       

/*** ...multiply, or just divide?... ***/
