/*
 * Copyright (C) Jason Gilbert <jason@scott.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * mrd - a simple program that prints out the remote commands as they occur
 */

#include <ggi/gii.h>

#include "buttonCodes.h"

static gii_input_t input;

#define EVT_MASK emPtrRelative | emPtrButton

void mrd_exit(int code)
{
   giiExit();
   exit(code);
}

void changeMode(int mode)
{
   printf("Current mode is %d.\n", mode);
}

void processRemoteButton(int num)
{
   printf("Button ");

   switch (num) {
      case CH_PLUS:
         printf("Channel Plus");
         break;
      case CH_MINUS:
         printf("Channel Minus");
         break;
      case CH_LAST:
         printf("Last Channel");
         break;
      case MODE_PC:
         printf("Mode PC");
         changeMode(num);
         break;
      case MODE_CD:
         printf("Mode CD");
         changeMode(num);
         break;
      case MODE_WEB:
         printf("Mode Web");
         changeMode(num);
         break;
      case MODE_DVD:
         printf("Mode DVD");
         changeMode(num);
         break;
      case MODE_PHONE:
         printf("Mode Phone");
         changeMode(num);
         break;
      case MUTE:
         printf("Mute");
         break;
      case AB:
         printf("A*B");
         break;
      case DISP:
         printf("Display");
         break;
      case GUIDE:
         printf("Guide");
         break;
      case VOL_PLUS:
         printf("Volume Plus");
         break;
      case VOL_MINUS:
         printf("Volume Minus");
         break;
      case POWER:
         printf("Power");
         break;
      case PLAY:
         printf("Play");
         break;
      case PAUSE:
         printf("Pause");
         break;
      case STOP:
         printf("Stop");
         break;
      case REW:
         printf("Rewind");
         break;
      case FF:
         printf("FastForward");
         break;
      case REC:
         printf("Record");
         break;
      case SELECT:
         printf("Select");
         break;
      case NUM_ENT:
         printf("Enter");
         break;
      case NUM_UP:
         printf("Up Arrow");
         break;
      default:
         if (num >= NUM_0 && num <= NUM_9) {
            printf("Number %d", num - 64);
         }
   }

   printf(" Pressed.\n");
}

void mouseMoved(gii_pmove_event ev)
{
   if (ev.y == -127) {
      processRemoteButton(ev.x);
   } else {
      printf("mouse moved: %d, %d\n", ev.x, ev.y);
   }
}

void mousePressed(gii_pbutton_event ev)
{
   printf("mouse pressed: %d\n", ev.button);
}

void mouseReleased(gii_pbutton_event ev)
{
   printf("mouse released: %d\n", ev.button);
}

int main()
{
   gii_event ev;

   if (giiInit()) {
      exit(1);
   }

   if ((input = giiOpen("linux-mouse", NULL)) == NULL) {
      mrd_exit(1);
   }

   if (giiSetEventMask(input, EVT_MASK)) {
      mrd_exit(1);
   }

   printf("waiting for events...\n");
   while (giiEventRead(input, &ev, EVT_MASK)) {
      /* printf("read event.\n");*/
      switch (ev.any.type) {
         case evPtrRelative:
            mouseMoved(ev.pmove);
            break;
         case evPtrButtonPress:
            mousePressed(ev.pbutton);
            break;
         case evPtrButtonRelease:
            mouseReleased(ev.pbutton);
            break;
         default:
            printf("Unknown event type %d\n", ev.any.type);
      }
   }

   giiClose(input);
   giiExit();

   printf("mrd success!\n");

   return 0;
}

