/*
 * Copyright (C) Jason Gilbert <jason@scott.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * xmms_mr.c - a XMMS plugin for using the X10 mouseRemote as a remote control.
 */

#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>

#include <ggi/gii.h>

#include <pthread.h>

#include <stdio.h>

#include "buttonCodes.h"

#define VERSION "0.1.0"

static gii_input_t input;

static pthread_t tid;

#define GII_EVT_MASK emPtrRelative

void* event_processor(void*);

void mr_init()
{
   pthread_attr_t attr;

   if (giiInit()) {
      g_warning("Failed to init GII library.");
      exit(1);
   }
   
   if ((input = giiOpen("linux-mouse", NULL)) == NULL) {
      g_warning("Failed to open mouse device.");
   }

   if (giiSetEventMask(input, GII_EVT_MASK)) {
      g_warning("Failed to set event mask.");
   }

   pthread_attr_init(&attr);
   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
   pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
   pthread_attr_setschedpolicy(&attr, SCHED_OTHER);
   pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

   pthread_create(&tid, &attr, event_processor, NULL);
}

void mr_about()
{
   g_warning("mouseRemote about()");
}

void mr_configure()
{
   g_warning("No mouseRemote configuration available.");
}

void mr_cleanup()
{
   pthread_cancel(tid);
   giiClose(input);
   giiExit();
}

GeneralPlugin mr_plugin = {
   NULL,
   NULL,
   -1,
   "X10 mouseRemote Controller Plugin "VERSION,
   mr_init,
   mr_about,
   mr_configure,
   mr_cleanup,
};

GeneralPlugin* get_gplugin_info()
{
   return &mr_plugin;
}

void* event_processor(void* args)
{
   gii_event evt;

   while(giiEventRead(input, &evt, GII_EVT_MASK)) {
      if (evt.any.type == evPtrRelative && evt.pmove.y == -127) {
         int vol = 0;
         switch (evt.pmove.x) {
            case PLAY:
               xmms_remote_play(mr_plugin.xmms_session);
               break;
            case STOP:
               xmms_remote_stop(mr_plugin.xmms_session);
               break;
            case PAUSE:
               xmms_remote_pause(mr_plugin.xmms_session);
               break;
            case CH_PLUS:
               /* actually forward track */
               xmms_remote_playlist_next(mr_plugin.xmms_session);
               break;
            case CH_MINUS:
               /* actually previous track */
               xmms_remote_playlist_prev(mr_plugin.xmms_session);
               break;
            case CH_LAST:
               xmms_remote_toggle_repeat(mr_plugin.xmms_session);
               break;
            case VOL_PLUS:
               vol = xmms_remote_get_main_volume(mr_plugin.xmms_session);
               if (vol < 100) {
                  xmms_remote_set_main_volume(mr_plugin.xmms_session, vol + 5);
               }
               break;
            case VOL_MINUS:
               vol = xmms_remote_get_main_volume(mr_plugin.xmms_session);
               if (vol > 0) {
                  xmms_remote_set_main_volume(mr_plugin.xmms_session, vol - 5);
               }
               break;
            case AB:
               /* A*B button above PLAY */
               xmms_remote_toggle_shuffle(mr_plugin.xmms_session);
               break;
            case POWER:
               break;
         }
      }
   }
}
